#include "module.h"
#include "redirfs/rfs.h"
#include "mnt_excludes.h"
#include <linux/namei.h>
#include <linux/list.h>
#include <linux/version.h>


struct kavoas_excluded
{
    struct vfsmount* mnt;
    char *mntpath;
    struct list_head list;
};

int interceptor_put_path(char const* mntfile);
int interceptor_get_path(char const* mntfile);

static LIST_HEAD(kavoas_excluded_list);
static DEFINE_MUTEX(kavoas_excluded_mutex);


static int kavoas_is_mnt_excluded_unsafe(struct vfsmount* vfs)
{
    struct kavoas_excluded* mnt;

    list_for_each_entry(mnt, &kavoas_excluded_list, list)
    {
        if (vfs == mnt->mnt)
        {
            return 1;
        }
    }

    return 0;
}

static void kavoas_excluded_remove(struct kavoas_excluded* excl)
{
    pr_debug("%s\n", __func__);

    list_del_init(&excl->list);
    kfree(excl->mntpath);
    kfree(excl);
}

int kavoas_mnt_excludes_add_path(char const* mntfile)
{
    struct vfsmount* vfs = NULL;
    int ret = 0;
    int added = 0;
    struct path mntpath;

    int badpath = kern_path(mntfile, LOOKUP_FOLLOW, &mntpath);

    pr_debug("%s: excluding %s\n", __func__, mntfile);

    if (!badpath && mntpath.mnt)
    {
        vfs = mntget(mntpath.mnt);
    }
    else
    {
        pr_err("Failed to exclude path %s\n", mntfile);
        ret = -1;
        goto release_path;
    }

    mutex_lock(&kavoas_excluded_mutex);
    if (!kavoas_is_mnt_excluded_unsafe(vfs))
    {
        struct kavoas_excluded* emnt;
        emnt = kmalloc(sizeof(*emnt), GFP_KERNEL);
        if (!emnt)
            goto nomem;

        emnt->mnt = vfs;
        emnt->mntpath = kstrdup(mntfile, GFP_KERNEL);

        if (!emnt->mntpath)
        {
            kfree(emnt);
            goto nomem;
        }
        list_add_tail(&emnt->list, &kavoas_excluded_list);
        added = 1;
    }
    mutex_unlock(&kavoas_excluded_mutex);

    if (vfs)
        vfs->mnt_flags |= MNT_SKIP_KAVOAS;

    if (added) {
        pr_debug("Adding to excludes path %s\n", mntfile);
        interceptor_put_path(mntfile);
    }

    goto exit;

nomem:
    mutex_unlock(&kavoas_excluded_mutex);
    pr_err("Failed to allocate memory for path %s\n", mntfile);
exit:
    mntput(vfs);
release_path:
    if (!badpath)
        path_release(&mntpath);

    return ret;
}

int kavoas_is_mnt_excluded(struct vfsmount* vfs)
{
    int found;

    pr_debug("%s: vfs = %p\n", __func__, vfs);

    mutex_lock(&kavoas_excluded_mutex);
    found = kavoas_is_mnt_excluded_unsafe(vfs);
    mutex_unlock(&kavoas_excluded_mutex);

    return found;
}

void kavoas_excluded_clean(void)
{
    pr_debug("%s\n", __func__);

    if (kavoas_rst_path("all"))
    {
        pr_err("Failed to clean kavoas mountpoint exclude list\n");
    }
}

static int kavoas_rst_mountpoint(const char* mntpath)
{
    struct vfsmount* vfs = NULL;
    struct path kernelpath;
    int badpath = kern_path(mntpath, LOOKUP_FOLLOW, &kernelpath);

    pr_debug("%s: %s\n", __func__, mntpath);
    if (badpath || !kernelpath.mnt)
    {
        pr_err("%s: Failed to find %s\n", __func__, mntpath);
        return -1;
    }

    vfs = mntget(kernelpath.mnt);
    vfs->mnt_flags &= ~MNT_SKIP_KAVOAS;
    mntput(vfs);

    if (!badpath)
        path_release(&kernelpath);

    return 0;
}

int kavoas_rst_path(char const* mntpath)
{
    struct kavoas_excluded *entry, *etmp;
    int all = 0;

    pr_debug("%s: %s\n", __func__, mntpath);

    if (strcmp(mntpath, "all") == 0)
    {
        all = 1;
    }

    mutex_lock(&kavoas_excluded_mutex);
    list_for_each_entry_safe(entry, etmp, &kavoas_excluded_list, list)
    {
        if (all || strcmp(entry->mntpath, mntpath) == 0)
        {
            kavoas_rst_mountpoint(entry->mntpath);
            kavoas_excluded_remove(entry);
        }
    }
    mutex_unlock(&kavoas_excluded_mutex);

    return 0;
}
