
/*
 * This source file is a part of a Kaspersky Antivirus Monitor.
 * Copyright (C) Kaspersky Lab, 1997-2010
 * See License.txt for details
 *
 */

#include "module.h"
#include "product_config.h"
#include "oas/small/protocol_alloc.h"
#include <linux/module.h>
#include <linux/init.h>
#include <linux/seq_file.h>

struct vfsmount* orig_rootmnt = NULL;
struct dentry* orig_root      = NULL;

/* Module section */
MODULE_AUTHOR("Kaspersky Lab");
MODULE_DESCRIPTION("Kaspersky Anti-Virus Monitor Module for Linux.");
#ifdef LINUX_KERNEL26
MODULE_INFO(supported, "external");

extern int use_binfmt;
module_param(use_binfmt, int, S_IRUGO);
MODULE_PARM_DESC(use_binfmt, "Enables hook on exec system call.");
#endif

static atomic_t monitor_count;

static int kav4fs_proc_version_print(struct seq_file* m, void* v)
{
    seq_printf(m, "%s", LFS_PRODUCT_VERSION);
    return 0;
}

static int kav4fs_version_open(struct inode* inode, struct file* file)
{
    return single_open(file, kav4fs_proc_version_print, NULL);
}

static const struct file_operations kav4fs_version_fops = {
    .owner                                        = THIS_MODULE,
    .open                                         = kav4fs_version_open,
    .read                                         = seq_read,
    .llseek                                       = seq_lseek,
    .release                                      = single_release,
};

void Monitor_ins_use(void)
{
#ifdef LINUX_KERNEL26
    try_module_get(THIS_MODULE);
#else
    MOD_INC_USE_COUNT;
#endif

    atomic_inc(&monitor_count);
}

void Monitor_dec_use(void)
{
#ifdef LINUX_KERNEL26
    module_put(THIS_MODULE);
#else
    MOD_DEC_USE_COUNT;
#endif

    atomic_dec(&monitor_count);
}

int Monitor_busy(void)
{
    return atomic_read(&monitor_count) ? 1 : 0;
}


static int __init kavmon_init_module(void)
{
    if (get_current_fs_root(&orig_rootmnt, &orig_root) < 0)
        return -1;

    atomic_xchg(&monitor_count, 0);

    if (!proc_create(KAV4FS_VERSION_NAME, 0, NULL, &kav4fs_version_fops))
    {
        printk(KERN_INFO "klflt: Warning! Failed to create %s\n",
               KAV4FS_VERSION_NAME);
        return -1;
    }

    return Monitor_interface_init();
}


static void __exit kavmon_cleanup_module(void)
{
    Monitor_interface_cleanup();

    dput(orig_root);
    mntput(orig_rootmnt);
    remove_proc_entry(KAV4FS_VERSION_NAME, NULL);
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 3, 48))
module_init(kavmon_init_module);
module_exit(kavmon_cleanup_module);
MODULE_VERSION(LFS_PRODUCT_VERSION);
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
#endif
